# -*- coding: utf-8 -*-
from functools import partial
import threading
import transaction
from zope.interface import implementer
from transaction.interfaces import IDataManager
from celery.app import app_or_default


__all__ = [
    'CeleryDataManager',
    'TransactionalTask',
    'task_tm',
]


# New Celery structure complicates all of this in the name of flexibility
Task = app_or_default().create_task_cls()
base_task = app_or_default().task

# Thread-local data
_thread_data = threading.local()


def _remove_manager(*args):
    try:
        del _thread_data.task_manager
    except AttributeError:
        pass


def _get_manager():
    task_manager = getattr(_thread_data, 'task_manager', None)
    if task_manager is None:
        task_manager = _thread_data.task_manager = CeleryDataManager()

    tx = transaction.get()
    tx.join(task_manager)
    tx.addAfterCommitHook(_remove_manager)
    return task_manager


@implementer(IDataManager)
class CeleryDataManager(object):
    transaction_manager = None

    def __init__(self):
        self.queued_tasks = []
        self.in_commit = False

    def _discard_tasks(self):
        """Discard all pending thread-local Celery tasks."""
        self.queued_tasks = []

    def _cleanup(self):
        self.queued_tasks = []

    def append(self, task):
        self.queued_tasks.append(task)

    def commit(self, trans):
        self.in_commit = True

    def sortKey(self):
        return str(id(self))

    def tpc_begin(self, transaction):
        pass

    def tpc_vote(self, trans):
        pass

    def tpc_finish(self, transaction):
        while self.queued_tasks:
            cls, args, kwargs = self.queued_tasks.pop(0)
            cls.original_apply_async(*args, **kwargs)
        self.in_commit = False
        self._cleanup()

    def tpc_abort(self, transaction):
        self._discard_tasks()
        self.in_commit = False

    abort = tpc_abort


class TransactionalTask(Task):
    """A task whose execution is delayed until after the current transaction.
    """
    abstract = True

    def original_apply_async(self, *args, **kwargs):
        """Shortcut method to reach real implementation of
        celery.Task.apply_sync"""
        return super(TransactionalTask, self).apply_async(*args, **kwargs)

    def apply_async(self, *args, **kwargs):
        _get_manager().append((self, args, kwargs))

task_tm = partial(base_task, base=TransactionalTask)



#!/usr/bin/env python
# -*- coding: utf-8 -*-


try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup


readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

requirements = [
    'transaction',
]

test_requirements = [
    # TODO: put package test requirements here
]

setup(
    name='pyramid_transactional_celery',
    version='0.1.1',
    description='A transaction-aware Celery job setup',
    long_description=readme + '\n\n' + history,
    author='Christopher Petrilli',
    author_email='petrilli@amber.org',
    url='https://github.com/petrilli/pyramid_transactional_celery',
    packages=[
        'pyramid_transactional_celery',
    ],
    package_dir={'pyramid_transactional_celery':
                 'pyramid_transactional_celery'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='pyramid_transactional_celery',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    test_suite='tests',
    tests_require=test_requirements
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_pyramid_transactional_celery
----------------------------------

Tests for `pyramid_transactional_celery` module.  In the end, we don't need
to test anything more than whether it properly handles the transactions, and
queues things at the right time.
"""

import unittest
import transaction
from uuid import uuid4
from celery import Celery


# XXX: Globals are evil. Need to find a better way to do this
task_output = []


class TestPyramidTransactionalCelery(unittest.TestCase):
    def setUp(self):
        self.celery_app = Celery()
        self.celery_app.conf.CELERY_ALWAYS_EAGER = True

        from pyramid_transactional_celery import (TransactionalTask, task_tm)

        class TestTask(TransactionalTask):
            """A simple task that appends a sentinel object to a global."""

            def run(self, sentinel):
                task_output.append(sentinel)

        @task_tm
        def test_task(sentinel):
            task_output.append(sentinel)

        self.celery_app.tasks['test_task_class'] = TestTask
        self.celery_app.tasks['test_task_fun'] = test_task
        transaction.begin()

    def test_successful_transaction(self):
        """Ensure that a successful transaction fires the task."""
        sentinel = uuid4()
        self.celery_app.tasks['test_task_class']().delay(sentinel)
        transaction.commit()
        self.assertIn(sentinel, task_output)

    def test_multiple_tasks(self):
        sentinels = [uuid4() for i in range(10)]
        for sentinel in sentinels:
            self.celery_app.tasks['test_task_class']().delay(sentinel)
        transaction.commit()
        self.assertListEqual(sentinels, task_output)

    def test_aborted_transaction(self):
        """Ensure that an aborted transaction does not fire the task."""
        sentinel = uuid4()
        self.celery_app.tasks['test_task_class']().delay(sentinel)
        transaction.abort()
        self.assertNotIn(sentinel, task_output)

    def test_successful_function_task(self):
        """Ensure that a successful transaction fires the task, even when it
        is a function."""
        sentinel = uuid4()
        self.celery_app.tasks['test_task_fun'].delay(sentinel)
        transaction.commit()
        self.assertIn(sentinel, task_output)

    def tearDown(self):
        global task_output
        task_output[:] = []

if __name__ == '__main__':
    unittest.main()
